@@ -344,6 +344,9 @@ static void *item_crawler_thread(void *arg) {
                     lru_crawler_class_done(i);
                     continue;
                 }
+            } else if (active_crawler_mod.mod->needs_client) {
+                lru_crawler_class_done(i);
+                continue;
             }
             pthread_mutex_lock(&lru_locks[i]);
             search = do_item_crawl_q((item *)&crawlers[i]);
@@ -397,8 +400,11 @@ static void *item_crawler_thread(void *arg) {
     if (active_crawler_mod.mod != NULL) {
         if (active_crawler_mod.mod->finalize != NULL)
             active_crawler_mod.mod->finalize(&active_crawler_mod);
-        if (active_crawler_mod.c.c != NULL) {
+        while (active_crawler_mod.c.c != NULL && bipbuf_used(active_crawler_mod.c.buf)) {
             lru_crawler_poll(&active_crawler_mod.c);
+        }
+        // Double checking in case the client closed during the poll
+        if (active_crawler_mod.c.c != NULL) {
             lru_crawler_release_client(&active_crawler_mod.c);
         }
         active_crawler_mod.mod = NULL;
@@ -584,7 +590,7 @@ enum crawler_result_type lru_crawler_crawl(char *slabs, const enum crawler_run_t
              p = strtok_r(NULL, ",", &b)) {
 
             if (!safe_strtoul(p, &sid) || sid < POWER_SMALLEST
-                    || sid >= MAX_NUMBER_OF_SLAB_CLASSES-1) {
+                    || sid >= MAX_NUMBER_OF_SLAB_CLASSES) {
                 pthread_mutex_unlock(&lru_crawler_lock);
                 return CRAWLER_BADCLASS;
             }
