@@ -30,6 +30,7 @@ typedef struct {
     uint64_t tailrepairs;
     uint64_t expired_unfetched; /* items reclaimed but never touched */
     uint64_t evicted_unfetched; /* items evicted but never touched */
+    uint64_t evicted_active; /* items evicted that should have been shuffled */
     uint64_t crawler_reclaimed;
     uint64_t crawler_items_checked;
     uint64_t lrutail_reflocked;
@@ -148,7 +149,8 @@ item *do_item_alloc(char *key, const size_t nkey, const unsigned int flags,
     uint8_t nsuffix;
     item *it = NULL;
     char suffix[40];
-    if (nbytes < 2 || nkey < 0)
+    // Avoid potential underflows.
+    if (nbytes < 2)
         return 0;
 
     size_t ntotal = item_make_header(nkey + 1, flags, nbytes, suffix, &nsuffix);
@@ -519,6 +521,7 @@ void item_stats_totals(ADD_STAT add_stats, void *c) {
             pthread_mutex_lock(&lru_locks[i]);
             totals.expired_unfetched += itemstats[i].expired_unfetched;
             totals.evicted_unfetched += itemstats[i].evicted_unfetched;
+            totals.evicted_active += itemstats[i].evicted_active;
             totals.evicted += itemstats[i].evicted;
             totals.reclaimed += itemstats[i].reclaimed;
             totals.crawler_reclaimed += itemstats[i].crawler_reclaimed;
@@ -535,6 +538,10 @@ void item_stats_totals(ADD_STAT add_stats, void *c) {
                 (unsigned long long)totals.expired_unfetched);
     APPEND_STAT("evicted_unfetched", "%llu",
                 (unsigned long long)totals.evicted_unfetched);
+    if (settings.lru_maintainer_thread) {
+        APPEND_STAT("evicted_active", "%llu",
+                    (unsigned long long)totals.evicted_active);
+    }
     APPEND_STAT("evictions", "%llu",
                 (unsigned long long)totals.evicted);
     APPEND_STAT("reclaimed", "%llu",
@@ -581,6 +588,7 @@ void item_stats(ADD_STAT add_stats, void *c) {
             totals.reclaimed += itemstats[i].reclaimed;
             totals.expired_unfetched += itemstats[i].expired_unfetched;
             totals.evicted_unfetched += itemstats[i].evicted_unfetched;
+            totals.evicted_active += itemstats[i].evicted_active;
             totals.crawler_reclaimed += itemstats[i].crawler_reclaimed;
             totals.crawler_items_checked += itemstats[i].crawler_items_checked;
             totals.lrutail_reflocked += itemstats[i].lrutail_reflocked;
@@ -624,6 +632,10 @@ void item_stats(ADD_STAT add_stats, void *c) {
                             "%llu", (unsigned long long)totals.expired_unfetched);
         APPEND_NUM_FMT_STAT(fmt, n, "evicted_unfetched",
                             "%llu", (unsigned long long)totals.evicted_unfetched);
+        if (settings.lru_maintainer_thread) {
+            APPEND_NUM_FMT_STAT(fmt, n, "evicted_active",
+                                "%llu", (unsigned long long)totals.evicted_active);
+        }
         APPEND_NUM_FMT_STAT(fmt, n, "crawler_reclaimed",
                             "%llu", (unsigned long long)totals.crawler_reclaimed);
         APPEND_NUM_FMT_STAT(fmt, n, "crawler_items_checked",
@@ -812,7 +824,8 @@ item *do_item_get(const char *key, const size_t nkey, const uint32_t hv, conn *c
     if (settings.verbose > 2)
         fprintf(stderr, "\n");
     /* For now this is in addition to the above verbose logging. */
-    LOGGER_LOG(c->thread->l, LOG_FETCHERS, LOGGER_ITEM_GET, NULL, was_found, key, nkey);
+    LOGGER_LOG(c->thread->l, LOG_FETCHERS, LOGGER_ITEM_GET, NULL, was_found, key, nkey,
+               (it) ? ITEM_clsid(it) : 0);
 
     return it;
 }
@@ -944,6 +957,9 @@ static int lru_pull_tail(const int orig_id, const int cur_lru,
                     if ((search->it_flags & ITEM_FETCHED) == 0) {
                         itemstats[id].evicted_unfetched++;
                     }
+                    if ((search->it_flags & ITEM_ACTIVE)) {
+                        itemstats[id].evicted_active++;
+                    }
                     LOGGER_LOG(NULL, LOG_EVICTIONS, LOGGER_EVICTION, search);
                     do_item_unlink_nolock(search, hv);
                     removed++;
@@ -1037,7 +1053,7 @@ static int lru_maintainer_juggle(const int slabs_clsid) {
  * The latter is to avoid newly started daemons from waiting too long before
  * retrying a crawl.
  */
-static void lru_maintainer_crawler_check(struct crawler_expired_data *cdata) {
+static void lru_maintainer_crawler_check(struct crawler_expired_data *cdata, logger *l) {
     int i;
     static rel_time_t next_crawls[MAX_NUMBER_OF_SLAB_CLASSES];
     static rel_time_t next_crawl_wait[MAX_NUMBER_OF_SLAB_CLASSES];
@@ -1060,13 +1076,9 @@ static void lru_maintainer_crawler_check(struct crawler_expired_data *cdata) {
             /* Need to think we can free at least 1% of the items before
              * crawling. */
             /* FIXME: Configurable? */
-            uint64_t low_watermark = (s->seen / 100) + 1;
+            uint64_t low_watermark = (possible_reclaims / 100) + 1;
             rel_time_t since_run = current_time - s->end_time;
             /* Don't bother if the payoff is too low. */
-            if (settings.verbose > 1)
-                fprintf(stderr, "maint crawler[%d]: low_watermark: %llu, possible_reclaims: %llu, since_run: %u\n",
-                        i, (unsigned long long)low_watermark, (unsigned long long)possible_reclaims,
-                        (unsigned int)since_run);
             for (x = 0; x < 60; x++) {
                 available_reclaims += s->histo[x];
                 if (available_reclaims > low_watermark) {
@@ -1088,9 +1100,13 @@ static void lru_maintainer_crawler_check(struct crawler_expired_data *cdata) {
             }
 
             next_crawls[i] = current_time + next_crawl_wait[i] + 5;
-            if (settings.verbose > 1)
-                fprintf(stderr, "maint crawler[%d]: next_crawl: %u, [%d] now: [%d]\n",
-                        i, next_crawl_wait[i], next_crawls[i], current_time);
+            LOGGER_LOG(l, LOG_SYSEVENTS, LOGGER_CRAWLER_STATUS, NULL, i, (unsigned long long)low_watermark,
+                    (unsigned long long)available_reclaims,
+                    (unsigned int)since_run,
+                    next_crawls[i] - current_time,
+                    s->end_time - s->start_time,
+                    s->seen,
+                    s->reclaimed);
             // Got our calculation, avoid running until next actual run.
             s->run_complete = false;
             pthread_mutex_unlock(&cdata->lock);
@@ -1119,6 +1135,11 @@ static void *lru_maintainer_thread(void *arg) {
     memset(&cdata, 0, sizeof(struct crawler_expired_data));
     pthread_mutex_init(&cdata.lock, NULL);
     cdata.crawl_complete = true; // kick off the crawler.
+    logger *l = logger_create();
+    if (l == NULL) {
+        fprintf(stderr, "Failed to allocate logger for LRU maintainer thread\n");
+        abort();
+    }
 
     pthread_mutex_lock(&lru_maintainer_lock);
     if (settings.verbose > 2)
@@ -1152,7 +1173,7 @@ static void *lru_maintainer_thread(void *arg) {
         }
         /* Once per second at most */
         if (settings.lru_crawler && last_crawler_check != current_time) {
-            lru_maintainer_crawler_check(&cdata);
+            lru_maintainer_crawler_check(&cdata, l);
             last_crawler_check = current_time;
         }
     }
@@ -1162,6 +1183,7 @@ static void *lru_maintainer_thread(void *arg) {
 
     return NULL;
 }
+
 int stop_lru_maintainer_thread(void) {
     int ret;
     pthread_mutex_lock(&lru_maintainer_lock);
