@@ -10,11 +10,15 @@
 #define LOGGER_ENTRY_MAX_SIZE 2048
 #define GET_LOGGER() ((logger *) pthread_getspecific(logger_key));
 
+/* Inlined from memcached.h - should go into sub header */
+typedef unsigned int rel_time_t;
+
 enum log_entry_type {
     LOGGER_ASCII_CMD = 0,
     LOGGER_EVICTION,
     LOGGER_ITEM_GET,
-    LOGGER_ITEM_STORE
+    LOGGER_ITEM_STORE,
+    LOGGER_CRAWLER_STATUS,
 };
 
 enum log_entry_subtype {
@@ -49,19 +53,23 @@ struct logentry_eviction {
     uint32_t latime;
     uint16_t it_flags;
     uint8_t nkey;
+    uint8_t clsid;
     char key[];
 };
 
 struct logentry_item_get {
     uint8_t was_found;
     uint8_t nkey;
+    uint8_t clsid;
     char key[];
 };
 
 struct logentry_item_store {
     int status;
     int cmd;
+    rel_time_t ttl;
     uint8_t nkey;
+    uint8_t clsid;
     char key[];
 };
 
